import java.awt.geom.QuadCurve2D;
import java.util.Iterator;

public class HashTable<E extends KeyedElementInterface<K>, K> implements HashTableInterface<E, K> {

	public static enum ProbeType {Linear, Quadratic, DoubleHashing}
	
	private static final int NUMBER_OF_BUCKETS = 17_321; // prime number
	
	private Bucket[] buckets;
	private int size;
	private ProbeType probeType = ProbeType.DoubleHashing;
	
	private int totalNumberOfProbes = 0;
	
	public ProbeType getProbeType() {
		return probeType;
	}

	public void setProbeType(ProbeType probeType) {
		this.probeType = probeType;
	}

	public int getTotalNumberOfProbes() {
		return totalNumberOfProbes;
	}

	public void setTotalNumberOfProbes(int totalNumberOfProbes) {
		this.totalNumberOfProbes = totalNumberOfProbes;
	}

	public HashTable() {
		this.size = 0;
		this.totalNumberOfProbes = 0;
		this.buckets = new Bucket[NUMBER_OF_BUCKETS];
		for ( int i = 0 ; i < this.buckets.length ; i++ ) {
			this.buckets[i] = new Bucket();
			this.buckets[i].element = null;
			this.buckets[i].emptyFromTheBeginning = true;
		}
	}
	
	
	
	@Override
	public Iterator<E> iterator() {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public int size() {
		return this.size;
	}

	@Override
	public boolean isEmpty() {
		return (this.size == 0);
	}

	@Override
	public HashTableInterface<E, K> copy() throws InstantiationException {
		// TODO Auto-generated method stub
		return null;
	}

	@Override
	public E insert(E element) {
		K key = element.getKey();
		int hashCode = key.hashCode() % this.buckets.length;
		int probes = 1;
		if (this.size == this.buckets.length) {
			// hash table is full, do replacement only
			while ((this.buckets[hashCode].element != null) && (probes <= this.buckets.length)) {
				if (key.equals(((E)(this.buckets[hashCode].element)).getKey())) {
					// this is a replacement of an existing element with same key
					E oldElement = (E) this.buckets[hashCode].element;
					this.buckets[hashCode].element = element;
					this.totalNumberOfProbes += probes;
					return oldElement;
				}
				
				// need to probe
				hashCode = this.probe(hashCode, probes++);
			}
			// no element with same key to replace, return null
			this.totalNumberOfProbes += probes;
			return null;
		} else {
			// add to hash table that still has room
			while (this.buckets[hashCode].element != null) {
				if (key.equals(((E)(this.buckets[hashCode].element)).getKey())) {
					// this is a replacement of an existing element with same key
					E oldElement = (E) this.buckets[hashCode].element;
					this.buckets[hashCode].element = element;
					return oldElement;
				}
				
				// need to probe
				hashCode = this.probe(hashCode, probes++);
			}
			
			// place element in bucket
			this.buckets[hashCode].element = element;
			this.buckets[hashCode].emptyFromTheBeginning = false;

			this.size++;
			this.totalNumberOfProbes += probes;
			return null;
		}
	}

	@Override
	public E get(K key) {
		int hashCode = key.hashCode() % this.buckets.length;
		int probes = 1;
		
		while ((probes <= this.buckets.length) && (!this.buckets[hashCode].emptyFromTheBeginning)) {
			if (this.buckets[hashCode].element == null) {
				// bucket is empty but not from the beginning
				hashCode = this.probe(hashCode, probes++);
			} else {
				// bucket has an element
				K otherKey = ((E) this.buckets[hashCode].element).getKey();
				if (key.equals(otherKey)) {
					// the bucket has element with same key, found it!
					this.totalNumberOfProbes += probes;
					return (E) this.buckets[hashCode].element;
				} else {
					// the bucket has element with different key
					hashCode = this.probe(hashCode, probes++);
				}
			}
		}
		
		this.totalNumberOfProbes += probes;
		return null;
	}

	@Override
	public E remove(K key) {
		int hashCode = key.hashCode() % this.buckets.length;
		int probes = 1;

		while ((probes <= this.buckets.length) && (!this.buckets[hashCode].emptyFromTheBeginning)) {
			if (this.buckets[hashCode].element == null) {
				// bucket is empty but not from the beginning
				hashCode = this.probe(hashCode, probes++);
			} else {
				// bucket has an element
				K otherKey = ((E) this.buckets[hashCode].element).getKey();
				if (key.equals(otherKey)) {
					// the bucket has element with same key, found it!
					this.totalNumberOfProbes += probes;
					E element = (E) this.buckets[hashCode].element;
					this.buckets[hashCode].element = null;
					this.size--;
					return element;
				} else {
					// the bucket has element with different key
					hashCode = this.probe(hashCode, probes++);
				}
			}
		}
		
		// element not found
		return null;
	}

	@Override
	public void removeAll() {
		this.size = 0;
		this.totalNumberOfProbes = 0;
		this.buckets = new Bucket[NUMBER_OF_BUCKETS];
		for ( int i = 0 ; i < this.buckets.length ; i++ ) {
			this.buckets[i] = new Bucket();
			this.buckets[i].element = null;
			this.buckets[i].emptyFromTheBeginning = true;
		}
	}

	@Override
	public int getSizeOfLargestBucket() {
		// TODO Auto-generated method stub
		return 0;
	}

	@Override
	public double getAverageBucketSize() {
		// TODO Auto-generated method stub
		return 0;
	}

	@Override
	public Object[] getBuckets() {
		// TODO Auto-generated method stub
		return null;
	}

	
	private int probe(int hashCode, int i) {
		switch (this.probeType) {
		case Linear:
			hashCode = (hashCode + 1) % this.buckets.length;
			break;
		case Quadratic:
			hashCode = (hashCode + i + (i * i)) % this.buckets.length;
			break;
		case DoubleHashing:
			hashCode = (hashCode + i + (int)(7919.0/953.0)*hashCode) % this.buckets.length;
			break;
		}
		return hashCode;
	}
	
	
}

	class Bucket<E> {
	public E element;
	public boolean emptyFromTheBeginning;	
}
